library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- Version 5
-- 100 x 32 character Display 
-- using 800 x 600 timings with 40mhz clock


entity gfx is
port (
clk		:	in std_logic ;	-- 40mhz master clock SVGA
reset		:	in	std_logic ;	-- system master reset
wr			:	in	std_logic ;	-- CPU Write
VBI		:	out	std_logic ;	-- interupt

-- memory interface
Vaddr		:	out std_logic_vector(11 downto 0);	-- vram address lines (4k DP ram)
Vdata		:	in std_logic_vector(7 downto 0);		-- video data
-- vga output
hsync		:	out std_logic;		-- VGA Syncs
vsync		:	out std_logic;
red		:	out std_logic;		-- 6 colour bit output
red_i		:	out std_logic;
red_x		:	out std_logic;    -- no extra bit for green
green		:	out std_logic;
green_i	:	out std_logic;
blue		:	out std_logic;
blue_i	:	out std_logic;
blue_x	:	out std_logic;
-- control input
option	:	in std_logic;	-- active low write to control regiser
Raddr		:	in std_logic_vector(1 downto 0);	-- up to 4 registers
Rdata		:	in std_logic_vector(7 downto 0)	-- 8 bit colour data from CPU
) ;
end entity ;
architecture behavioral of gfx is
constant H_Porch:    integer := 800;
constant H_Sync:     integer := 840;
constant H_SyncEnd:  integer := 968;
constant H_read:		integer := 1048;
constant H_Total:    integer := 1056;

signal H_Count:      integer range 0 to H_Total;
signal H_Blank:      std_logic := '0';

constant B_Border:   integer := 576;
constant V_Porch:    integer := 600;
constant V_Sync:     integer := 601;
constant V_SyncEnd:  integer := 605;
constant V_Total:    integer := 628;

signal V_Count:      integer range 0 to V_Total;
signal V_Blank:      std_logic := '0';
signal V_Border:     std_logic := '0';

constant CharHeight: integer := 8-1;
constant RamSize:    integer := 4096-1;
signal RamPointer: 	integer := 4096-1;
signal C_count:    	integer := 4096-1;

signal char_data:		std_logic_vector (7 downto 0);
signal pixel_data:	std_logic_vector (7 downto 0);
signal ink_data:		std_logic_vector (7 downto 0);
signal paper_data:	std_logic_vector (7 downto 0);
signal border_data:	std_logic_vector (7 downto 0);

signal regUpdate:			std_logic :='1' ;
signal NMIenable:			std_logic :='1' ;
signal blank_area:		std_logic :='0' ;
signal border_area:		std_logic :='0' ;
signal Scanline:			integer range 0 to 17 ;	-- 9 row charcter spacing
signal sequence:			integer range 0 to 7 ;	-- 8 step memory timing

begin
-- define the screen area's, sysn is wholy within blank. blank is wholy within border
-- anything not border is active.
blank_area   <= '0' when v_blank ='0' or h_blank ='0' else '1';
border_area  <= '0' when  v_border ='0' or Scanline>15 or blank_a rea = '0' else '1';
regUpdate	<= '0' when option='0' and wr='0' else '1';

MemoryControl:
Process(clk,H_count)
begin
	IF rising_edge(clk) then
		if (h_count = h_read) then       -- first memory access starts 6 clocks before the active area
			c_count <= ramPointer ;   
				sequence <= 0 ;
		end if ;
			if (h_count< h_porch) or (h_count > h_read) then
			case sequence is
				when 0 =>		-- put the character cell address onto the address bus
					Vaddr(11 downto 0) <= std_logic_vector(to_unsigned(c_count,12));
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ; 
					sequence <= 1;
				when 1 =>
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					sequence <= 2;
				when 2 =>
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					char_data <= vdata ;
					sequence <= 3;
				when 3 =>		-- put the character data address ontot he bus
					Vaddr(11 downto 7) <= char_data(4 downto 0);
					Vaddr(6 downto 5) <= "11" ;
					Vaddr(4 downto 3) <= char_data(6 downto 5);
					Vaddr(2 downto 0) <= std_logic_vector(to_unsigned(Scanline/2,3));
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					sequence <= 4;
					c_count <= c_count+1 ;		-- character count has done it's job and can be updated
				when 4 =>
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					sequence <= 5;
				when 5 =>
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					sequence <= 6;
				when 6 =>
					pixel_data(7 downto 1) <= pixel_data(6 downto 0) ;
					sequence <= 7;
				when 7 =>		-- read the pixel data from the chatater map
					if char_data(7) ='0' then
						pixel_data <= vdata ;
					else
						pixel_data <= not(vdata) ;
					end if ;
					sequence <= 0;
			end case ;	
		end if ;
	end if;
end process ;

configuartion:
process(reset,regUpdate)
begin
	if reset='0' then
		ink_data    <= "11111111" ;
		paper_data  <= "11010110" ;
		border_data <= "00101001" ;	
		NMIenable <='1' ;
	elsif rising_edge(regUpdate) then
		case Raddr is			
		when "00" =>
			ink_data 	<= Rdata ;
		when "01" =>
			paper_data 	<= Rdata ;
		when "10" =>
			border_data <= Rdata ;
		when "11" =>
			NMIenable	<= not(Rdata(0)) ;
		end case ;
	end if ;
end process;

-- shift register runs on the opposite edge of the clock to the memory
pixelOutput:
process(clk)
begin
	if falling_edge(clk) then
		if (border_area='1') then	
			if pixel_data(7)='1' then
				red_x 	<= ink_data(0) ;
				red_i 	<= ink_data(1) ;
				red 		<= ink_data(2) ;
				green_i 	<= ink_data(3) ;
				green 	<= ink_data(4) ;
				blue_x 	<= ink_data(5) ;
				blue_i 	<= ink_data(6) ;
				blue 		<= ink_data(7) ;		
			else
				red_x 	<= paper_data(0) ;
				red_i 	<= paper_data(1) ;
				red 		<= paper_data(2) ;
				green_i 	<= paper_data(3) ;
				green 	<= paper_data(4) ;
				blue_x 	<= paper_data(5) ;
				blue_i 	<= paper_data(6) ;
				blue 		<= paper_data(7) ;
			end if;
		else	
			if blank_area ='0' then
				red_x 	<= '0' ;
				red_i 	<= '0' ;
				red 		<= '0' ;
				green_i 	<= '0' ;
				green 	<= '0' ;
				blue_x 	<= '0' ;
				blue_i 	<= '0' ;
				blue 		<= '0' ;
			else
				red_x 	<= border_data(0) ;
				red_i 	<= border_data(1) ;
				red 		<= border_data(2) ;
				green_i 	<= border_data(3) ;
				green 	<= border_data(4) ;
				blue_x	<= border_data(5) ;
				blue_i	<= border_data(6) ;
				blue	 	<= border_data(7) ;
			end if ;
		end if;
	end if;
end process;

-- Horizontal Timing

-- Master horizontal counter, in character positions.
-- H_LineEnd is set for the last character position in the row, when this falls
-- again at the start of the next row, that triggers the vertical display actions 

HorizontalCount:
process(Clk, reset)
begin
    if (reset = '0') then
        H_Count <= 0;
    elsif rising_edge(Clk) then
        if H_Count = (H_Total - 1) then
            H_Count <= 0;
        else
             H_Count <= H_Count + 1;
        end if;
    end if;
end process;

-- The display is "active" up to the start of the front porch, after that is
-- either sync or porch until the other border starts, for the the output multiplexer
-- all non displayed area is the same, so only a generic "blank" is needed.
-- active low 

HorizontalDisplay:
process(Clk)
begin
    if rising_edge(Clk) then
        if (H_Count >= H_Porch) then
            H_Blank <= '0';
        else
            H_Blank <= '1';
        end if;
    end if;
end process;


-- Horizontal sync is active in the blank part of the display
-- It has it's own output, so won't interfere with the blanking signal
-- positive sysnce on 800x 600
HorizontalSync:
process(clk)
begin
    if rising_edge(clk) then
        if ((H_Count >= H_Sync) and (H_Count< H_SyncEnd)) then
            HSYNC <= '1';
        else
            HSYNC <= '0';
        end if;
    end if;
end process;

--
-- Vertical Timing
-- similar to horizontal timing in that there are 4 types of display
-- sync, blank, border and active

-- Master vertical counter, counting in rows triggered by the line end pulse

VerticalCount:
process (clk, reset)
begin
    if (reset = '0') then
			V_Count <= 0;
			rampointer <= 0;
			Scanline <= 0;
    elsif rising_edge(clk) then
		if (h_count = h_sync) then
        if (V_Count = (V_Total-1)) then
            V_Count <= 0;
				rampointer <= 0;
				Scanline <= 0;
       else
            V_Count <= V_Count + 1;
				Scanline <= Scanline+1 ;
        end if;
		  if Scanline = 17 then
				Scanline <= 0 ;
				ramPointer <= RamPointer+128 ;
			end if;
       end if;
    end if;
end process;

-- Similar to the horizontal counter, the display is active from row 0, up to
-- the start of the porch area, where the blanking signal is set until the
-- border re-starts
-- active low

VerticalDisplay:
process(clk)
begin
    if rising_edge(clk) then
        if (V_Count >= H_Porch)  then
            V_Blank <= '0';
        else
            V_Blank <= '1';
        end if;
    end if;
end process;

-- Vertical sync is active for 2 full rows on the middle of the blanking period
-- positive sysnce on 800x 600

VerticalSync:
process(clk)
begin
    if rising_edge(clk) then
        if ((V_Count>= V_Sync) and (V_Count < V_SyncEnd) ) then
            VSYNC <= '1';
      else
            VSYNC <= '0';
        end if;
    end if;
end process;

verticalBorder:
process(clk)
begin
    if rising_edge(clk) then
        if V_Count >= B_border then
            v_border <= '0';
        else
            v_border <= '1';
        end if;
    end if;
end process;

interupt:
process(clk)
begin
	if rising_edge(clk) then
		if (V_Count = V_porch) and (H_count> H_sync) and (H_count < H_syncEnd) then
			vbi <= NMIenable ;
		else
			vbi <= '1' ;
		end if;
	end if;
end process;


end behavioral;
